import time

import napari
import numpy as np
import cv2
from napari import Viewer
from napari.layers import Labels
from scipy.interpolate import interp1d
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLabel, QSpinBox, QPushButton, QComboBox
from qtpy.QtCore import QThread, QObject, Signal
from scipy.ndimage import distance_transform_edt
from skimage.measure import regionprops
from skimage.morphology import binary_erosion
from skimage.transform import SimilarityTransform, warp
from ._utils import ProgressWidget
from ..mean_contour import settings
from ..mean_contour.meanContour import MeanThread
from ..mean_contour._contour import calcRpsvInterpolation
from ..mean_contour._reconstruction import reconstruct
from ..mean_contour._essentials import magnitude

import matplotlib.pyplot as plt

DISTANCE_BASED = "Distance-based"
CONTOUR_BASED = "Contour-based"
RPSV = "RPSV"


def contour_cv2_mask_uniform(mask, contoursize_max):
    mask = mask.astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    areas = [cv2.contourArea(cnt) for cnt in contours]
    max_ind = np.argmax(areas)
    contour = np.squeeze(contours[max_ind])
    contour = np.reshape(contour, (-1, 2))
    contour = np.append(contour, contour[0].reshape((-1, 2)), axis=0)
    contour = contour.astype('float32')

    rows, cols = mask.shape
    delta = np.diff(contour, axis=0)
    s = [0]
    for d in delta:
        dl = s[-1] + np.linalg.norm(d)
        s.append(dl)

    if (s[-1] == 0):
        s[-1] = 1

    s = np.array(s) / s[-1]
    fx = interp1d(s, contour[:, 0] / rows, kind='linear')
    fy = interp1d(s, contour[:, 1] / cols, kind='linear')
    S = np.linspace(0, 1, contoursize_max, endpoint=False)
    X = rows * fx(S)
    Y = cols * fy(S)

    contour = np.transpose(np.stack([X, Y])).astype(np.float32)

    contour = np.stack((contour[:, 1], contour[:, 0]), axis=-1)
    return contour


def average_mask(m1, m2, w1, w2):
    m1 = m1.astype(np.uint16)
    m2 = m2.astype(np.uint16)
    im_center = np.asarray(m1.shape)/2
    centroid1 = np.asarray(regionprops(m1)[0].centroid)
    centroid2 = np.asarray(regionprops(m2)[0].centroid)
    transl_1 = centroid1 - im_center
    transl_2 = centroid2 - im_center
    tform1 = SimilarityTransform(translation=np.flip(transl_1))
    tform2 = SimilarityTransform(translation=np.flip(transl_2))
    m1_translated = warp(m1, tform1, preserve_range=True).astype(bool)
    m2_translated = warp(m2, tform2, preserve_range=True).astype(bool)
    dt1_translated = distance_transform_edt(binary_erosion(m1_translated)) - distance_transform_edt(~m1_translated)
    dt2_translated = distance_transform_edt(binary_erosion(m2_translated)) - distance_transform_edt(~m2_translated)
    average_dist_translated = (w1 * dt1_translated + w2 * dt2_translated) / (w1+w2)
    average_mask_translated = average_dist_translated > 0
    transl_avg = im_center - (centroid1*w1+centroid2*w2)/(w1+w2)
    tform_avg = SimilarityTransform(translation=np.flip(transl_avg))

    return warp(average_mask_translated, tform_avg)


class InterpolationWorker(QObject):
    done = Signal("PyQt_PyObject")
    progress = Signal(int)
    dimension: int
    n_contour_points: int
    data: np.ndarray
    method: str
    max_iterations: int
    dims_displayed: tuple
    current_step: tuple
    selected_label: int

    def run(self):
        try:
            dimension = self.dimension
            n_contour_points = self.n_contour_points
            data = self.data.copy()
            method = self.method
            layer_slice_template = [
                slice(None) if d in self.dims_displayed
                    else None if d == dimension
                    else self.current_step[d]
                for d in range(data.ndim)]
            prev_cnt = None
            prev_layer = None
            prev_mask = None
            start = time.time()
            for i in range(data.shape[dimension]):
                self.progress.emit(i)
                layer_slice = layer_slice_template.copy()
                layer_slice[dimension] = i
                cur_mask = data[tuple(layer_slice)] == self.selected_label
                cur_mask = cur_mask.astype(np.uint8)
                if cur_mask.max() == 0:
                    continue
                next_layer_slice = layer_slice.copy()
                next_layer_slice[dimension] = i + 1
                prev_layer_slice = layer_slice.copy()
                prev_layer_slice[dimension] = i - 1
                if i + 1 < data.shape[dimension] \
                        and (data[tuple(prev_layer_slice)] == self.selected_label).max() > 0\
                        and (data[tuple(next_layer_slice)] == self.selected_label).max() > 0:
                    continue
                cnt = contour_cv2_mask_uniform(cur_mask, n_contour_points)
                centroid = cnt.mean(0)
                start_index = np.argmin(np.abs(np.arctan2(*(cnt - centroid).T)))
                cnt = np.roll(cnt, -start_index, 0)
                if prev_cnt is not None:
                    if method == RPSV:
                        stgs = settings.Settings(max_iterations=self.max_iterations,
                                                 n_points=n_contour_points)
                        rpsv_thread = MeanThread([prev_cnt, cnt], stgs)
                        rpsv_thread.run()
                    for j in range(prev_layer + 1, i):
                        inter_layer_slice = layer_slice_template.copy()
                        inter_layer_slice[dimension] = j
                        prev_w = i - j
                        cur_w = j - prev_layer
                        weights = [prev_w, cur_w]
                        if method == RPSV:
                            contours = rpsv_thread.contours
                            regularMean = np.zeros_like(contours[0].lookup[contours[0].parameterization, :])
                            for j in range(2):
                                regularMean += contours[j].lookup[contours[j].parameterization, :] * weights[j]
                            regularMean /= np.sum(weights)
                            q_mean = calcRpsvInterpolation(contours, weights)
                            guessRayLengths = np.zeros(contours[0].lookup[contours[0].parameterization].shape[0])
                            for i_contour in range(2):
                                contourtmp = contours[i_contour].lookup[contours[i_contour].parameterization]
                                contourlengths = magnitude(contourtmp)
                                guessRayLengths += contourlengths * weights[i_contour]
                            guessRayLengths /= np.sum(weights)
                            guessRayLengths = magnitude(regularMean)

                            qraylengths = magnitude(q_mean)
                            qraylengths[qraylengths < 1e-99] = 1e-99

                            dirs = q_mean / qraylengths.reshape(qraylengths.shape[0], 1)
                            r_mean_lengths, costs = reconstruct(q_mean, guessRayLengths.copy(), stgs, rpsv_thread.rpSignal)
                            mean_cnt = dirs * r_mean_lengths.reshape(r_mean_lengths.shape[0], 1)
                            mean_cnt = mean_cnt.astype(np.int32)
                            mask = np.zeros_like(data[tuple(inter_layer_slice)])
                            cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(self.selected_label), -1)
                        elif method == CONTOUR_BASED:
                            mean_cnt = (prev_w * prev_cnt + cur_w * cnt)/(prev_w + cur_w)
                            mean_cnt = mean_cnt.astype(np.int32)
                            mask = np.zeros_like(data[tuple(inter_layer_slice)])
                            cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(self.selected_label), -1)
                        elif method == DISTANCE_BASED:
                            mask = average_mask(prev_mask, cur_mask, prev_w, cur_w)
                            mask = mask.astype(np.uint8)
                            mask[mask > 0] = self.selected_label
                        else:
                            raise ValueError("method should be one of %s" % ((RPSV, CONTOUR_BASED, DISTANCE_BASED),))
                        cur_slice = data[tuple(inter_layer_slice)]
                        cur_slice[mask > 0] = mask[mask > 0]
                prev_cnt = cnt
                prev_layer = i
                prev_mask = cur_mask
            napari.notification_manager.receive_info("Done in %.4f s" % (time.time() - start))
            self.done.emit(data)
        except Exception as e:
            self.done.emit(None)
            raise e


class InterpolationWidget(QWidget):
    def __init__(self, viewer: Viewer):
        super().__init__()
        self.viewer = viewer
        self.progress_dialog = ProgressWidget(self, message="Interpolating slices...")
        layout = QVBoxLayout()
        self._active_labels_layer = None

        self.interpolation_thread = QThread()
        self.interpolation_worker = InterpolationWorker()
        self.interpolation_worker.moveToThread(self.interpolation_thread)
        self.interpolation_worker.done.connect(self.interpolation_thread.quit)
        self.interpolation_worker.done.connect(lambda _: self.progress_dialog.setVisible(False))
        self.interpolation_worker.done.connect(self.set_labels)
        self.interpolation_worker.progress.connect(self.progress_dialog.setValue)
        self.interpolation_thread.started.connect(self.interpolation_worker.run)
        layout.addWidget(QLabel("dimension"))
        self.dimension_dropdown = QSpinBox()
        self.dimension_dropdown.setMinimum(0)
        self.dimension_dropdown.setMaximum(0)
        self.dimension_dropdown.setToolTip("Dimension along which slices will be interpolated")
        layout.addWidget(self.dimension_dropdown)

        layout.addWidget(QLabel("Method"))
        self.method_dropdown = QComboBox()
        self.method_dropdown.addItem(CONTOUR_BASED)
        self.method_dropdown.addItem(DISTANCE_BASED)
        self.method_dropdown.addItem(RPSV)
        self.method_dropdown.setToolTip("Interpolation method\n"
                                        "%s: arithmetic mean of contour points\n"
                                        "%s: interpolation between distance maps\n"
                                        "%s: shape-aware mean of contours" % (CONTOUR_BASED, DISTANCE_BASED, RPSV))
        self.method_dropdown.currentTextChanged.connect(lambda _: self.rpsv_widget.setVisible(self.method_dropdown.currentText() == "RPSV"))
        layout.addWidget(self.method_dropdown)

        layout.addWidget(QLabel("# contour points"))
        self.n_points = QSpinBox()
        self.n_points.setMinimum(10)
        self.n_points.setMaximum(1000)
        self.n_points.setValue(300)
        self.n_points.setToolTip("Number of contour points sampled. Used only for\"%s\" and \"%s\" methods" % (RPSV, CONTOUR_BASED))
        layout.addWidget(self.n_points)

        self.rpsv_widget = QWidget()
        rpsv_layout = QVBoxLayout()
        rpsv_layout.addWidget(QLabel("max iterations"))
        self.rpsv_iterations_spinbox = QSpinBox()
        self.rpsv_iterations_spinbox.setMaximum(100)
        self.rpsv_iterations_spinbox.setMinimum(1)
        self.rpsv_iterations_spinbox.setValue(20)
        self.rpsv_iterations_spinbox.setToolTip("Maximum number of iterations for RPSV. Can be fewer if points converge.")
        rpsv_layout.addWidget(self.rpsv_iterations_spinbox)
        self.rpsv_widget.setLayout(rpsv_layout)
        self.rpsv_widget.setVisible(self.method_dropdown.currentText() == RPSV)
        rpsv_layout.setContentsMargins(0, 0, 0, 0)
        self.rpsv_widget.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.rpsv_widget)

        self.interpolate_button = QPushButton("Interpolate")
        self.interpolate_button.clicked.connect(self.interpolate)
        self.interpolation_worker.done.connect(lambda: self.interpolate_button.setEnabled(True))
        layout.addWidget(self.interpolate_button)
        self.viewer.layers.selection.events.connect(self.on_layer_selection_change)
        viewer.dims.events.order.connect(self.on_order_change)
        viewer.dims.events.ndisplay.connect(lambda _: self.interpolate_button.setEnabled(viewer.dims.ndisplay == 2))
        self.setLayout(layout)
        layout.addStretch()
        self.on_layer_selection_change()

    def on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            active_layer = self.viewer.layers.selection.active
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer

    @property
    def active_labels_layer(self):
        return self._active_labels_layer

    @active_labels_layer.setter
    def active_labels_layer(self, layer):
        if layer is None:
            return
        self.dimension_dropdown.setMaximum(layer.ndim)
        if self._active_labels_layer is not None:
            self._active_labels_layer.bind_key("Control-I", overwrite=True)(None)
        self._active_labels_layer = layer
        layer.bind_key("Control-I", overwrite=True)(self.interpolate)
        self.on_order_change()

    def on_order_change(self, event=None):
        if self.active_labels_layer is None or self.active_labels_layer.ndim < 3 or self.viewer.dims.ndisplay == 3:
            return
        new_dim = self.viewer.dims.not_displayed[0]
        self.dimension_dropdown.setValue(new_dim)

    def prepare_interpolation_worker(self):
        self.interpolation_worker.dimension = self.dimension_dropdown.value()
        self.interpolation_worker.n_contour_points = self.n_points.value()
        self.interpolation_worker.data = self.active_labels_layer.data
        self.interpolation_worker.method = self.method_dropdown.currentText()
        self.interpolation_worker.dims_displayed = self.viewer.dims.displayed
        self.interpolation_worker.current_step = self.viewer.dims.current_step
        self.interpolation_worker.selected_label = self.active_labels_layer.selected_label
        self.interpolation_worker.max_iterations = self.rpsv_iterations_spinbox.value()

    def interpolate(self, _=None):
        if self.active_labels_layer is None:
            return
        self.progress_dialog.setMaximum(self.active_labels_layer.data.shape[self.dimension_dropdown.value()])
        self.progress_dialog.setVisible(True)
        self.interpolate_button.setEnabled(False)
        self.prepare_interpolation_worker()
        self.interpolation_thread.start()

    def set_labels(self, data):
        if data is None:
            return
        update_mask = data > 0
        if self.active_labels_layer.preserve_labels:
            update_mask &= self.active_labels_layer.data == 0
        self.active_labels_layer.data[update_mask] = data[update_mask]
        self.active_labels_layer.events.data()
        self.active_labels_layer.refresh()



import numpy as np
cimport numpy as np
import time
from enum import Enum
from cython.parallel cimport prange
cimport cython
from libc.math cimport sqrt
class ReconstructionMethods(Enum):
    NEWTON = 1
    GRADIENT_DESCENT = 2
    CG = 3
    SKIP = 4
    JOZSI_GRADIENT = 5
    

# derivative approximation
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef np.ndarray[np.double_t, ndim=2] dt(np.ndarray[np.double_t, ndim=2] points, int order):
    cdef int n_points = points.shape[0]
    cdef double next_x_1, next_y_1, next_x_2, next_y_2, next_x_3, next_y_3, next_x_4, next_y_4
    cdef double prev_x_1, prev_y_1, prev_x_2, prev_y_2, prev_x_3, prev_y_3, prev_x_4, prev_y_4
    cdef double prev_factor_x, prev_factor_y, next_factor_x, next_factor_y, curr_factor_x, curr_factor_y
    cdef np.ndarray[np.double_t, ndim=2] out = np.empty_like(points)
    cdef int i
    for i in prange(n_points, nogil=True):
        if i < 4:
            prev_x_4 = points[n_points + i - 4, 0]
            prev_y_4 = points[n_points + i - 4, 1]
            if i < 3:
                prev_x_3 = points[n_points + i - 3, 0]
                prev_y_3 = points[n_points + i - 3, 1]
                if i < 2:
                    prev_x_2 = points[n_points + i - 2, 0]
                    prev_y_2 = points[n_points + i - 2, 1]
                    if i < 1:
                        prev_x_1 = points[n_points + i - 1, 0]
                        prev_y_1 = points[n_points + i - 1, 1]
                    else:
                        prev_x_1 = points[i - 1, 0]
                        prev_y_1 = points[i - 1, 1]
                else:
                    prev_x_2 = points[i - 2, 0]
                    prev_y_2 = points[i - 2, 1]
                    prev_x_1 = points[i - 1, 0]
                    prev_y_1 = points[i - 1, 1]
            else:
                prev_x_3 = points[i - 3, 0]
                prev_y_3 = points[i - 3, 1]
                prev_x_2 = points[i - 2, 0]
                prev_y_2 = points[i - 2, 1]
                prev_x_1 = points[i - 1, 0]
                prev_y_1 = points[i - 1, 1]
        else:
            prev_x_4 = points[i - 4, 0]
            prev_y_4 = points[i - 4, 1]
            prev_x_3 = points[i - 3, 0]
            prev_y_3 = points[i - 3, 1]
            prev_x_2 = points[i - 2, 0]
            prev_y_2 = points[i - 2, 1]
            prev_x_1 = points[i - 1, 0]
            prev_y_1 = points[i - 1, 1]
        if i + 4 >= n_points:
            next_x_4 = points[i+4-n_points, 0]
            next_y_4 = points[i+4-n_points, 1]
            if i + 3 >= n_points:
                next_x_3 = points[i + 3 - n_points, 0]
                next_y_3 = points[i + 3 - n_points, 1]
                if i + 2 >= n_points:
                    next_x_2 = points[i + 2 - n_points, 0]
                    next_y_2 = points[i + 2 - n_points, 1]
                    if i + 1 >= n_points:
                        next_x_1 = points[i + 1 - n_points, 0]
                        next_y_1 = points[i + 1 - n_points, 1]
                    else:
                        next_x_1 = points[i + 1, 0]
                        next_y_1 = points[i + 1, 1]
                else:
                    next_x_2 = points[i + 2, 0]
                    next_y_2 = points[i + 2, 1]
                    next_x_1 = points[i + 1, 0]
                    next_y_1 = points[i + 1, 1]
            else:
                next_x_3 = points[i + 3, 0]
                next_y_3 = points[i + 3, 1]
                next_x_2 = points[i + 2, 0]
                next_y_2 = points[i + 2, 1]
                next_x_1 = points[i + 1, 0]
                next_y_1 = points[i + 1, 1]
        else:
            next_x_4 = points[i+4, 0]
            next_y_4 = points[i+4, 1]
            next_x_3 = points[i + 3, 0]
            next_y_3 = points[i + 3, 1]
            next_x_2 = points[i + 2, 0]
            next_y_2 = points[i + 2, 1]
            next_x_1 = points[i + 1, 0]
            next_y_1 = points[i + 1, 1]
        if order == 1:
            prev_factor_x = (1. / 280.) * prev_x_4 - (4. / 105.) * prev_x_3 + (1. / 5.) * prev_x_2 - (4. / 5.) * prev_x_1
            prev_factor_y = (1. / 280.) * prev_y_4 - (4. / 105.) * prev_y_3 + (1. / 5.) * prev_y_2 - (4. / 5.) * prev_y_1
            next_factor_x = (1. / 280.) * next_x_4 - (4. / 105.) * next_x_3 + (1. / 5.) * next_x_2 - (4. / 5.) * next_x_1
            next_factor_y = (1. / 280.) * next_y_4 - (4. / 105.) * next_y_3 + (1. / 5.) * next_y_2 - (4. / 5.) * next_y_1
            out[i, 0] = prev_factor_x+next_factor_x
            out[i, 1] = prev_factor_y+next_factor_y
        elif order == 2:
            prev_factor_x = (-1./560.)*prev_x_4+(8./315.)*prev_x_3-(1./5.)*prev_x_2+(8./5.)*prev_x_1
            prev_factor_y = (-1./560.)*prev_y_4+(8./315.)*prev_y_3-(1./5.)*prev_y_2+(8./5.)*prev_y_1
            next_factor_x = (-1. / 560.) * next_x_4 + (8. / 315.) * next_x_3 - (1. / 5.) * next_x_2 + (8. / 5.) * next_x_1
            next_factor_y = (-1. / 560.) * next_y_4 + (8. / 315.) * next_y_3 - (1. / 5.) * next_y_2 + (8. / 5.) * next_y_1
            curr_factor_x = (-205. / 72.) * points[i, 0]
            curr_factor_y = (-205. / 72.) * points[i, 1]
            out[i, 0] = prev_factor_x + next_factor_x + curr_factor_x
            out[i, 1] = prev_factor_y + next_factor_y + curr_factor_y
    return out

# magnitude of a single vector or a set of vectors
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef magnitude(np.ndarray[np.double_t, ndim=2] points):
    cdef np.ndarray[np.double_t, ndim=1] out = np.empty(points.shape[0], np.float64)
    cdef int i
    for i in prange(points.shape[0], nogil=True):
        out[i] = sqrt(points[i, 0]*points[i, 0] + points[i, 1]*points[i, 1])
    return out
    # return np.sqrt(points[:,0]*points[:,0]+points[:,1]*points[:,1])
        
# inner product of two curves
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray[np.double_t, ndim=1] innerProduct(np.ndarray[np.double_t, ndim=2] curve1, np.ndarray[np.double_t, ndim=2] curve2):
    # return np.sum(curve1*curve2, axis=1)
    return curve1[:,0]*curve2[:,0]+curve1[:,1]*curve2[:,1]

class Contour:

    def __init__(self, pointArray, nPoi, resMultiplier):
        # number of points in contour
        self.nPoi = pointArray.shape[0]
        self.cPoints = nPoi
        self.resMultiplier = resMultiplier
        self.resolution = self.cPoints*self.resMultiplier
        self.pointArray = pointArray
        self.contourLength = self.getContourLength()
        self.centroid = self.getCentroid()
        self.lookup, self.parameterization = self.getLookupTables()
        self.smoothLookupTable()
        self.diffs = np.empty(self.cPoints)
        self.calcParams()

    def getCentroid(self):
        centroid = np.zeros((1,2))
        for p in self.pointArray:
            centroid = centroid + p
        centroid = centroid/self.nPoi
        return centroid

    def dt(self, order):
        '''nextParam = self.nextParam
        prevParam = self.prevParam
        nextParam2 = np.roll(nextParam,-1,axis=0)
        prevParam2 = np.roll(prevParam,1,axis=0)'''

        if order==1:
            return dt(self.lookup[self.parameterization], 1)
            #return (1/12)*self.lookup[prevParam2]-(2/3)*self.lookup[prevParam]+(2/3)*self.lookup[nextParam]-(1/12)*self.lookup[nextParam]
            #return (self.lookup[nextParam,:]-self.lookup[prevParam,:])/2
        if order==2:
            return dt(self.lookup[self.parameterization], 2)
            #return (-1/12)*self.lookup[prevParam2]+(4/3)*self.lookup[prevParam]-(5/2)*self.lookup[self.parameterization]+(4/3)*self.lookup[nextParam]-(1/12)*self.lookup[nextParam2]
            #return self.lookup[nextParam,:]+self.lookup[prevParam,:]-2*self.lookup[self.parameterization,:]
    
    
    #Christoffel divergence as defined in (9): Gamma_i
    def getChristoffelDivergence(self):
        deriv = self.derivatives
        sderiv = self.sderivatives
        return (deriv[:,0]*sderiv[:,0]+deriv[:,1]*sderiv[:,1])/(deriv[:,0]*deriv[:,0]+deriv[:,1]*deriv[:,1])

    def getSRV(self):
        deriv = self.derivatives
        return np.sqrt(magnitude(deriv)) 

    def getRPSV(self):
        return self.lookup[self.parameterization]*self.srv[:,None]

    def getSRVF(self):
        res = self.derivatives.copy()
        deriv = self.derivatives
        velo = magnitude(deriv)
        srv = np.sqrt(velo)
        res[:,0] /= srv
        res[:,1] /= srv
        return res

    def calcParams(self):
        
        self.nextParam = np.roll(self.parameterization, -1, axis=0)
        self.prevParam = np.roll(self.parameterization, 1, axis=0)

        self.derivatives = self.dt(1)
        self.sderivatives = self.dt(2)

        deriv = self.derivatives
        sderiv = self.sderivatives

        self.christoffel = self.getChristoffelDivergence()
        self.srv = self.getSRV()


    def getIdxDiff(self, dgamma):
        nextParam = self.nextParam
        prevParam = self.prevParam
        criteria1 = dgamma<0
        criteria2 = dgamma>=0

        diffs = self.diffs

        diffs[:] = 0
        
        diffs[criteria1] = self.parameterization[criteria1] - prevParam[criteria1]
        diffCriteria = diffs<0
        diffs[diffCriteria] = self.parameterization[diffCriteria]+self.cPoints*self.resMultiplier-prevParam[diffCriteria]

        diffs[criteria2] = nextParam[criteria2]-self.parameterization[criteria2]
        diffCriteria = diffs<0
        diffs[diffCriteria] = nextParam[diffCriteria]+self.cPoints*self.resMultiplier-self.parameterization[diffCriteria]
        

        return diffs

    # make sure we remain in lookup table index range with the reparameterization
    def getInRangeParameterization(self):
        pointNum = self.lookup.shape[0]
        criteria1 = self.parameterization.copy() >= pointNum
        criteria2 = self.parameterization.copy() < 0
        self.parameterization[criteria2] %= pointNum
        self.parameterization[criteria2] = -1*(self.parameterization[criteria2]-pointNum)
        self.parameterization[criteria1] %= pointNum

    # smoothing to ensure that 
    def smoothParameterization(self):
        resHalf = self.parameterization.shape[0]*self.resMultiplier/2

        tmparam = self.parameterization.copy()
        tmp = self.parameterization.copy()
        tmpnext = np.roll(tmp, -1, axis=0)
        tmpprev = np.roll(tmp, 1, axis=0)
        crit1 = (tmpnext < resHalf) & (tmp > resHalf) & (tmpprev > resHalf)
        tmpnext[crit1] += self.cPoints*self.resMultiplier
        crit2 = (tmp < resHalf) & (tmpprev > resHalf)
        tmp[crit2] += self.cPoints*self.resMultiplier
        tmpnext[crit2] += self.cPoints*self.resMultiplier

        self.parameterization = 2+2*tmp+tmpnext+tmpprev
        self.parameterization >>= 2

        self.getInRangeParameterization()

    def smoothLookupTable(self):
        nextArr = np.roll(self.lookup, -1, axis=0)
        prevArr = np.roll(self.lookup, 1, axis=0)

        temp = 0.25*(2* self.lookup + prevArr + nextArr)
        tempNext = np.roll(temp, -1, axis=0)
        tempPrev = np.roll(temp, 1, axis=0)

        self.lookup = 0.25*(2* temp + tempPrev + tempNext)

    def getContourLength(self):
        nextArr = np.roll(self.pointArray, -1, axis=0)

        cLength = np.sum(np.sqrt((nextArr[:,0]-self.pointArray[:,0])*(nextArr[:,0]-self.pointArray[:,0])+(nextArr[:,1]-self.pointArray[:,1])*(nextArr[:,1]-self.pointArray[:,1])))
        return cLength
    

    def isClockwise(self):
        nextLook = np.roll(self.lookup, -1, axis=0)
        edges = self.lookup[self.parameterization,0]*nextLook[self.parameterization,1]-nextLook[self.parameterization,0]*self.lookup[self.parameterization,1]
        if np.sum(edges)>0:
            return True
        else:
            return False

    def setStartingPointToLowestY(self):
        lowestY = np.argmin(self.lookup[:,1])
        self.lookup = np.roll(self.lookup, -1*lowestY, axis=0)

    def getLookupTables(self):
        start = time.time()
        # length of one step if we want to achieve self.resolution:
        unitLength = self.contourLength/self.resolution
        # lookup table
        lut = np.empty((2*self.resolution,2))
        # shifted point arrays
        nextArray = np.roll(self.pointArray, -1, axis=0)

        j = 0 # index of LookUpTable
        remainder = 0 # length of overflow for sections between 2 points
        sum_while = 0
        for i in range(self.nPoi):
            startPoint = self.pointArray[i]
            nextPoint = nextArray[i]
            direction = nextPoint-startPoint
            dirLen = np.sqrt(direction[0]*direction[0]+direction[1]*direction[1])
            direction /= dirLen # normalized direction between 2 points
            reqUnits = int(np.round(dirLen/unitLength))
            xcoords = np.linspace(startPoint[0], nextPoint[0], num=reqUnits)
            ycoords = np.linspace(startPoint[1], nextPoint[1], num=reqUnits)
            lut[j:(j+reqUnits),0] = xcoords
            lut[j:(j+reqUnits),1] = ycoords
            direction *= unitLength # set the length of the direction vector to contour unit length
            j += reqUnits
        
        lut_res = np.array(lut[0:j,:])
        #idxes = np.linspace(0,j-1, num=j).astype(int)
        #lut_idx = idxes[0:j:int(self.resolution/self.cPoints)]


        
        lut_idx = np.empty(self.cPoints, dtype=int)
        j = 0
        for i in range(0, self.resolution, int(self.resolution/self.cPoints)):
            lut_idx[j] = i
            j += 1
        return lut_res, lut_idx
    
    def export(self, filename):
        with open (filename, 'w') as expfile:
            for i in range(self.nPoi):
                expfile.write(str(self.pointArray[i,0])+","+str(self.pointArray[i,1])+"\n")
            expfile.close()




import numpy as np
import matplotlib.pyplot as plt
import math
import time
from enum import Enum
from scipy.interpolate import splprep, splev
class ReconstructionMethods(Enum):
    NEWTON = 1
    GRADIENT_DESCENT = 2
    CG = 3
    SKIP = 4
    JOZSI_GRADIENT = 5


def getCoefficientsForAccuracyLevel(aLevel, order):
    switcher1 = {
        1: [-1/2, 0, 1/2],
        2: [1/12, -2/3, 0, 2/3, -1/12],
        3: [-1/60, 3/20, -3/4, 0, 3/4, -3/20, 1/60],
        4: [1/280, -4/105, 1/5, -4/5, 0, 4/5, -1/5, 4/105, -1/280]
    }
    switcher2 = {
        1: [1, -2, 1],
        2: [-1/12, 4/3, -5/2, 4/3, -1/12],
        3: [1/90, -3/20, 3/2, -49/18, 3/2, -3/20, 1/90],
        4: [-1/560, 8/315, -1/5, 8/5, -205/72, 8/5, -1/5, 8/315, -1/560]
    }

    if order==1:
        return switcher1.get(aLevel)
    if order==2:
        return switcher2.get(aLevel)
    

# derivative approximation
def dt(points, order):

    pNext = np.roll(points, -1, axis=0)
    pPrev = np.roll(points, 1, axis=0)

    pNext2 = np.roll(points, -2, axis=0)
    pPrev2 = np.roll(points, 2, axis=0)

    pNext3 = np.roll(points, -3, axis=0)
    pPrev3 = np.roll(points, 3, axis=0)

    pNext4 = np.roll(points, -4, axis=0)
    pPrev4 = np.roll(points, 4, axis=0)
    #d = points-pPrev
    #e = pNext-points
    #d_abs = magnitude(d).reshape(d.shape[0],1)
    #e_abs = magnitude(e).reshape(e.shape[0],1)

    if order==1:
        prevFactors = (1/280)*pPrev4-(4/105)*pPrev3+(1/5)*pPrev2-(4/5)*pPrev
        nextFactors = -(1/280)*pNext4+(4/105)*pNext3-(1/5)*pNext2+(4/5)*pNext
        return prevFactors+nextFactors
        #retval = ((d_abs*e_abs)/(d_abs+e_abs))*(d/d_abs**2 + e/e_abs**2)
        #return retval/magnitude(retval).reshape(d.shape[0],1)
        #return (pNext-pPrev)/2
        #return (1/12)*pPrev2-(2/3)*pPrev+(2/3)*pNext-(1/12)*pNext2

    if order==2:
        prevFactors = (-1/560)*pPrev4+(8/315)*pPrev3-(1/5)*pPrev2+(8/5)*pPrev
        currFactor = (-205/72)*points
        nextFactors = (-1/560)*pNext4+(8/315)*pNext3-(1/5)*pNext2+(8/5)*pNext
        return prevFactors+currFactor+nextFactors
        #retval = (2/(d_abs+e_abs))*(e/e_abs - d/d_abs)
        #return retval/magnitude(retval).reshape(d.shape[0],1)
        #return pNext+pPrev-2*points
        #return (-1/12)*pPrev2+(4/3)*pPrev-(5/2)*points+(4/3)*pNext-(1/12)*pNext2


# magnitude of a single vector or a set of vectors
def magnitude(points):
    if len(points.shape)>1:
        return np.sqrt(points[:,0]*points[:,0]+points[:,1]*points[:,1])
    else:
        return np.sqrt(points[0]*points[0]+points[1]*points[1])
        
# inner product of two curves
def innerProduct(curve1,curve2):
    return np.sum(curve1*curve2, axis=1)
    # return curve1[:,0]*curve2[:,0]+curve1[:,1]*curve2[:,1]

class Contour:

    def __init__(self, pointArray, nPoi, resMultiplier):
        # number of points in contour
        self.nPoi = pointArray.shape[0]
        self.cPoints = nPoi
        self.resMultiplier = resMultiplier
        self.resolution = self.cPoints*self.resMultiplier
        self.pointArray = pointArray
        self.contourLength = self.getContourLength()
        self.centroid = self.getCentroid()
        self.lookup, self.parameterization = self.getLookupTables()
        self.smoothLookupTable()
        self.diffs = np.empty(self.cPoints)
        self.calcParams()

    def getCentroid(self):
        centroid = np.zeros((1,2))
        for p in self.pointArray:
            centroid = centroid + p
        centroid = centroid/self.nPoi
        return centroid

    def dt(self, order):
        '''nextParam = self.nextParam
        prevParam = self.prevParam
        nextParam2 = np.roll(nextParam,-1,axis=0)
        prevParam2 = np.roll(prevParam,1,axis=0)'''

        if order==1:
            return dt(self.lookup[self.parameterization], 1)
            #return (1/12)*self.lookup[prevParam2]-(2/3)*self.lookup[prevParam]+(2/3)*self.lookup[nextParam]-(1/12)*self.lookup[nextParam]
            #return (self.lookup[nextParam,:]-self.lookup[prevParam,:])/2
        if order==2:
            return dt(self.lookup[self.parameterization], 2)
            #return (-1/12)*self.lookup[prevParam2]+(4/3)*self.lookup[prevParam]-(5/2)*self.lookup[self.parameterization]+(4/3)*self.lookup[nextParam]-(1/12)*self.lookup[nextParam2]
            #return self.lookup[nextParam,:]+self.lookup[prevParam,:]-2*self.lookup[self.parameterization,:]
    
    
    #Christoffel divergence as defined in (9): Gamma_i
    def getChristoffelDivergence(self):
        deriv = self.derivatives
        sderiv = self.sderivatives
        return (deriv[:,0]*sderiv[:,0]+deriv[:,1]*sderiv[:,1])/(deriv[:,0]*deriv[:,0]+deriv[:,1]*deriv[:,1])

    def getSRV(self):
        deriv = self.derivatives
        return np.sqrt(magnitude(deriv)) 

    def getRPSV(self):
        return self.lookup[self.parameterization]*self.srv[:,None]

    def getSRVF(self):
        res = self.derivatives.copy()
        deriv = self.derivatives
        velo = magnitude(deriv)
        srv = np.sqrt(velo)
        res[:,0] /= srv
        res[:,1] /= srv
        return res

    def calcParams(self):
        
        self.nextParam = np.roll(self.parameterization, -1, axis=0)
        self.prevParam = np.roll(self.parameterization, 1, axis=0)

        self.derivatives = self.dt(1)
        self.sderivatives = self.dt(2)

        deriv = self.derivatives
        sderiv = self.sderivatives

        self.christoffel = self.getChristoffelDivergence()
        self.srv = self.getSRV()


    def getIdxDiff(self, dgamma):
        nextParam = self.nextParam
        prevParam = self.prevParam
        criteria1 = dgamma<0
        criteria2 = dgamma>=0

        diffs = self.diffs

        diffs[:] = 0
        
        diffs[criteria1] = self.parameterization[criteria1] - prevParam[criteria1]
        diffCriteria = diffs<0
        diffs[diffCriteria] = self.parameterization[diffCriteria]+self.cPoints*self.resMultiplier-prevParam[diffCriteria]

        diffs[criteria2] = nextParam[criteria2]-self.parameterization[criteria2]
        diffCriteria = diffs<0
        diffs[diffCriteria] = nextParam[diffCriteria]+self.cPoints*self.resMultiplier-self.parameterization[diffCriteria]
        

        return diffs

    # make sure we remain in lookup table index range with the reparameterization
    def getInRangeParameterization(self):
        pointNum = self.lookup.shape[0]
        criteria1 = self.parameterization.copy() >= pointNum
        criteria2 = self.parameterization.copy() < 0
        self.parameterization[criteria2] %= pointNum
        self.parameterization[criteria2] = -1*(self.parameterization[criteria2]-pointNum)
        self.parameterization[criteria1] %= pointNum

    # smoothing to ensure that 
    def smoothParameterization(self):
        resHalf = self.parameterization.shape[0]*self.resMultiplier/2

        tmparam = self.parameterization.copy()
        tmp = self.parameterization.copy()
        tmpnext = np.roll(tmp, -1, axis=0)
        tmpprev = np.roll(tmp, 1, axis=0)
        crit1 = (tmpnext < resHalf) & (tmp > resHalf) & (tmpprev > resHalf)
        tmpnext[crit1] += self.cPoints*self.resMultiplier
        crit2 = (tmp < resHalf) & (tmpprev > resHalf)
        tmp[crit2] += self.cPoints*self.resMultiplier
        tmpnext[crit2] += self.cPoints*self.resMultiplier

        self.parameterization = 2+2*tmp+tmpnext+tmpprev
        self.parameterization >>= 2

        self.getInRangeParameterization()

    def smoothLookupTable(self):
        nextArr = np.roll(self.lookup, -1, axis=0)
        prevArr = np.roll(self.lookup, 1, axis=0)

        temp = 0.25*(2* self.lookup + prevArr + nextArr)
        tempNext = np.roll(temp, -1, axis=0)
        tempPrev = np.roll(temp, 1, axis=0)

        self.lookup = 0.25*(2* temp + tempPrev + tempNext)

    def getContourLength(self):
        nextArr = np.roll(self.pointArray, -1, axis=0)

        cLength = np.sum(np.sqrt((nextArr[:,0]-self.pointArray[:,0])*(nextArr[:,0]-self.pointArray[:,0])+(nextArr[:,1]-self.pointArray[:,1])*(nextArr[:,1]-self.pointArray[:,1])))
        return cLength
    

    def isClockwise(self):
        nextLook = np.roll(self.lookup, -1, axis=0)
        edges = self.lookup[self.parameterization,0]*nextLook[self.parameterization,1]-nextLook[self.parameterization,0]*self.lookup[self.parameterization,1]
        if np.sum(edges)>0:
            return True
        else:
            return False

    def setStartingPointToLowestY(self):
        lowestY = np.argmin(self.lookup[:,1])
        self.lookup = np.roll(self.lookup, -1*lowestY, axis=0)

    def getLookupTables(self):
        start = time.time()
        # length of one step if we want to achieve self.resolution:
        unitLength = self.contourLength/self.resolution
        #print("unit length: "+str(unitLength))
        # lookup table
        lut = np.empty((2*self.resolution,2))
        # shifted point arrays
        nextArray = np.roll(self.pointArray, -1, axis=0)

        j = 0 # index of LookUpTable
        remainder = 0 # length of overflow for sections between 2 points
        sum_while = 0
        for i in range(self.nPoi):
            startPoint = self.pointArray[i]
            nextPoint = nextArray[i]
            direction = nextPoint-startPoint
            dirLen = np.sqrt(direction[0]*direction[0]+direction[1]*direction[1])
            #print("dirlen: "+str(dirLen))
            direction /= dirLen # normalized direction between 2 points
            reqUnits = int(np.round(dirLen/unitLength))
            xcoords = np.linspace(startPoint[0], nextPoint[0], num=reqUnits)
            ycoords = np.linspace(startPoint[1], nextPoint[1], num=reqUnits)
            lut[j:(j+reqUnits),0] = xcoords
            lut[j:(j+reqUnits),1] = ycoords
            direction *= unitLength # set the length of the direction vector to contour unit length
            j += reqUnits
        
        lut_res = np.array(lut[0:j,:])
        #idxes = np.linspace(0,j-1, num=j).astype(int)
        #lut_idx = idxes[0:j:int(self.resolution/self.cPoints)]


        
        lut_idx = np.empty(self.cPoints, dtype=int)
        j = 0
        for i in range(0, self.resolution, int(self.resolution/self.cPoints)):
            lut_idx[j] = i
            j += 1
        #print(xcoords)
        #print(j)
        return lut_res, lut_idx
    
    def export(self, filename):
        with open (filename, 'w') as expfile:
            for i in range(self.nPoi):
                expfile.write(str(self.pointArray[i,0])+","+str(self.pointArray[i,1])+"\n")
            expfile.close()




import scipy.interpolate as interp
class InterpHelper:
    def __init__(self):
        pass
    def setInterpolator(self,xpoints,ypoints):
        self.interpolator = interp.interp1d(x=xpoints,y=ypoints,kind='quadratic')



import numpy as np

def rk4(dydt, t0, y0, Q, Q_dot, theta, theta_dot, h=1, num_points=5):
    y_prev = y0
    t_prev = t0

    h # step size: this will determine how frequent the sampling should be in the function: for big h, the sampling is small
    num_points # this determines how many points should be "interpolated". does not influence sampling frequency, just the number of sampling points.
    y = [y0] # function to be determined: we only know its value in 1 point
    t = [t0]
    debug_vars = []

    for n in range(num_points):
        debug_vars.append(np.tan(y_prev))

        k1 = dydt(t_prev,y_prev, Q, Q_dot, theta, theta_dot)
        k2 = dydt(int(t_prev+h/2), y_prev+h*(k1/2), Q, Q_dot, theta, theta_dot)
        k3 = dydt(int(t_prev+h/2), y_prev+h*(k2/2), Q, Q_dot, theta, theta_dot)
        k4 = dydt(t_prev+h, y_prev+h*k3, Q, Q_dot, theta, theta_dot)

        qp = Q[t_prev]
        qdp = Q_dot[t_prev]
        tp = theta_dot[t_prev]
        tdp = theta_dot[t_prev]

        qn = Q[int(t_prev+h/2)]
        qdn = Q_dot[int(t_prev+h/2)]
        tn = theta_dot[int(t_prev+h/2)]
        tdn = theta_dot[int(t_prev+h/2)]

        y_next = y_prev+(1/6)*h*(k1+2*k2+2*k3+k4)

        tanyprev = np.tan(y_prev)
        tanynext = np.tan(y_next)

        #if np.sin(y_next)<0:
        #    y_next = 2*np.pi-y_next
        t_next = t_prev+h

        r_next = np.cbrt( np.sin(y_next)* Q[int(t_prev+h/2)]/theta_dot[int(t_prev+h/2)] )

        y.append(y_next)
        t.append(t_next)

        y_prev = y_next
        t_prev = t_next
    
    debug_vars.append(0)
    return t, y, debug_vars



from qtpy.QtWidgets import QWidget, QProgressBar, QLabel, QVBoxLayout
from qtpy.QtCore import Qt
from qtpy import QtGui

class ProgressWidget(QWidget):
    def __init__(self, parent=None, min_value=0, max_value=0, value=0, message="running...", visible=False, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setAttribute(Qt.WA_ShowWithoutActivating)
        self.drag_start_x = None
        self.drag_start_y = None
        layout = QVBoxLayout()
        self.label = QLabel(message)
        layout.addWidget(self.label)
        self.progress_bar = QProgressBar()
        self.progress_bar.setMinimum(min_value)
        self.progress_bar.setMaximum(max_value)
        self.progress_bar.setValue(value)
        self.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.CustomizeWindowHint)
        self.setVisible(visible)
        layout.addWidget(self.progress_bar)
        layout.addStretch()
        self.setLayout(layout)

    def reset(self):
        self.progress_bar.reset()

    def setMinimum(self, value):
        self.progress_bar.setMinimum(value)

    def setMaximum(self, value):
        self.progress_bar.setMaximum(value)

    def setValue(self, value):
        self.progress_bar.setValue(value)

    def setText(self, text):
        self.label.setText(text)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        self.drag_start_x = event.x()
        self.drag_start_y = event.y()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        self.move(event.globalX() - self.drag_start_x, event.globalY()-self.drag_start_y)